@@ -105,12 +105,14 @@ typedef enum {
     OPTION_HW_CURSOR,
     OPTION_NOACCEL,
     OPTION_ACCEL_METHOD,
+    OPTION_FP_DITHER,
 } G80Opts;
 
 static const OptionInfoRec G80Options[] = {
     { OPTION_HW_CURSOR,         "HWCursor",     OPTV_BOOLEAN,   {0}, FALSE },
     { OPTION_NOACCEL,           "NoAccel",      OPTV_BOOLEAN,   {0}, FALSE },
     { OPTION_ACCEL_METHOD,      "AccelMethod",  OPTV_STRING,    {0}, FALSE },
+    { OPTION_FP_DITHER,         "FPDither",     OPTV_BOOLEAN,   {0}, FALSE },
     { -1,                       NULL,           OPTV_NONE,      {0}, FALSE }
 };
 
@@ -313,6 +315,8 @@ G80PreInit(ScrnInfoPtr pScrn, int flags)
         goto fail;
     }
 
+    pNv->Dither = xf86ReturnOptValBool(pNv->Options, OPTION_FP_DITHER, FALSE);
+
     /* Set the bits per RGB for 8bpp mode */
     if(pScrn->depth == 8)
         pScrn->rgbBits = 8;
@@ -769,7 +773,7 @@ G80ScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
     ScrnInfoPtr pScrn;
     G80Ptr pNv;
     CARD32 pitch;
-    int visualMask;
+    int visualMask, i;
     BoxRec AvailFBArea;
 
     /* First get the ScrnInfoRec */
@@ -829,6 +833,10 @@ G80ScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
     AvailFBArea.y2 = pNv->offscreenHeight;
     xf86InitFBManager(pScreen, &AvailFBArea);
 
+    pNv->reg[0x00001708/4] = 0;
+    for(i = 0; i < 8; i++)
+        pNv->reg[0x00001900/4 + i] = 0;
+
     if(!pNv->NoAccel) {
         G80InitHW(pScrn);
         switch(pNv->AccelMethod) {
